#!/usr/bin/env python

"""
mtpy/utils/calculator.py

Helper functions for standard calculations, e.g. error propagation


@UofA, 2013
(LK)

"""

#=================================================================


import numpy as np
import math, cmath

import pycsamt.utils.exceptions as MTex


#=================================================================

#define uncertainty for differences between time steps
epsilon = 1e-9
#magnetic permeability in free space in H/m (=Vs/Am)
mu0 = 4e-7*math.pi


#=================================================================




def centre_point(xarray, yarray):
    """
    get the centre point of arrays of x and y values
    """
    return (xarray.max() + xarray.min())/2., (yarray.max() + yarray.min())/2.


def roundsf(number, sf):
    """
    round a number to a specified number of significant figures (sf)
    """
    # can't have < 1 s.f.
    sf = max(sf,1.)
    rounding = int(np.ceil(-np.log10(number) + sf - 1.))
    
    return np.round(number, rounding)


def get_period_list(period_min,period_max,periods_per_decade,include_outside_range=True):
    """
    get a list of values (e.g. periods), evenly spaced in log space and 
    including values on multiples of 10
    
    :returns:
        numpy array containing list of values
    
    :inputs:
        period_min = minimum period
        period_max = maximum period
        periods_per_decade = number of periods per decade
        include_outside_range = option whether to start and finish the period
                                list just inside or just outside the bounds
                                specified by period_min and period_max
                                default True
    
    """
    
    
    log_period_min = np.log10(period_min)
    log_period_max = np.log10(period_max)
    
    # check if log_period_min is a whole number
    if log_period_min % 1 > 0:
        # list of periods, around the minimum period, that will be present in specified 
        # periods per decade
        aligned_logperiods_min = np.linspace(np.floor(log_period_min),np.ceil(log_period_min),periods_per_decade + 1)
        lpmin_diff = log_period_min - aligned_logperiods_min
        # index of starting period, smallest value > 0
        if include_outside_range:
            spimin = np.where(lpmin_diff > 0)[0][-1]
        else:
            spimin = np.where(lpmin_diff < 0)[0][0]
        start_period = aligned_logperiods_min[spimin]
    else:
        start_period = log_period_min
    
    if log_period_max % 1 > 0:
        # list of periods, around the maximum period, that will be present in specified 
        # periods per decade
        aligned_logperiods_max = np.linspace(np.floor(log_period_max),np.ceil(log_period_max),periods_per_decade + 1)
        lpmax_diff = log_period_max - aligned_logperiods_max
        # index of starting period, smallest value > 0
        if include_outside_range:
            spimax = np.where(lpmax_diff < 0)[0][0]
        else:
            spimax = np.where(lpmax_diff > 0)[0][-1]
        stop_period = aligned_logperiods_max[spimax]
    else:
        stop_period = log_period_max        
        
    return np.logspace(start_period,stop_period,int((stop_period-start_period)*periods_per_decade + 1))


def nearest_index(val,array):
    """
    find the index of the nearest value in the array
    :param val: the value to search for
    :param array: the array to search in
    
    :return: index: integer describing position of nearest value in array
    
    """
    # absolute difference between value and array
    diff = np.abs(array-val)
    
    return np.where(diff==min(diff))[0][0]


def make_log_increasing_array(z1_layer, target_depth, n_layers, increment_factor=0.999):
    """
    create depth array with log increasing cells, down to target depth,
    inputs are z1_layer thickness, target depth, number of layers (n_layers)
    """

    # make initial guess for maximum cell thickness
    max_cell_thickness = target_depth
    # make initial guess for log_z
    log_z = np.logspace(np.log10(z1_layer),
                        np.log10(max_cell_thickness),
                        num=n_layers)
    counter = 0

    while np.sum(log_z) > target_depth:
        max_cell_thickness *= increment_factor
        log_z = np.logspace(np.log10(z1_layer),
                            np.log10(max_cell_thickness),
                            num=n_layers)
        counter += 1
        if counter > 1e6:
            break

    return log_z


def invertmatrix_incl_errors(inmatrix, inmatrix_err=None):

    if inmatrix is None:
        raise MTex.MTpyError_inputarguments('Matrix must be defined')

    if (inmatrix_err is not None) and (inmatrix.shape != inmatrix_err.shape):
        raise MTex.MTpyError_inputarguments('Matrix and err-matrix shapes do not match: %s - %s'%(str(inmatrix.shape), str(inmatrix_err.shape)))

    if (inmatrix.shape[-2] != inmatrix.shape[-1]):
        raise MTex.MTpyError_inputarguments('Matrices must be square!')

    if (inmatrix_err is not None) and (inmatrix_err.shape[-2] != inmatrix_err.shape[-1]) :
        raise MTex.MTpyError_inputarguments('Matrices must be square!')

    dim = inmatrix.shape[-1]


    det = np.linalg.det(inmatrix)

    if det == 0:
        raise MTex.MTpyError_inputarguments('Matrix is singular - I cannot invert that!')

    inv_matrix = np.zeros_like(inmatrix)



    if dim != 2:
        raise MTex.MTpyError_inputarguments('Only 2D matrices supported yet')

    inv_matrix = np.linalg.inv(inmatrix)

    inv_matrix_err = None

    if (inmatrix_err is not None):
        inmatrix_err = np.real(inmatrix_err)
        inv_matrix_err = np.zeros_like(inmatrix_err)

        for i in range(2):
            for j in range(2):
                #looping over the entries of the error matrix
                err = 0.
                for k in range(2):
                    for l in range(2):
                        #each entry has 4 summands

                        err += np.abs (- inv_matrix[i,k]  * inv_matrix[l,j]  * inmatrix_err[k,l])


                inv_matrix_err[i,j] = err

 
    return inv_matrix, inv_matrix_err

def rhophi2z(rho, phi, freq):
    """
        Convert impedance-style information given in Rho/Phi format into complex valued Z.

        Input:
        rho - 2x2 array (real) - in Ohm m
        phi - 2x2 array (real) - in degrees
        freq - scalar - frequency in Hz

        Output:
        Z - 2x2 array (complex)
    """

    try:
        if rho.shape != (2,2) or phi.shape != (2,2):
            raise
        if not (rho.dtype in ['float', 'int'] and phi.dtype in ['float', 'int']):
            raise

    except: 
        raise MTex.MTpyError_inputarguments('ERROR - arguments must be two 2x2 arrays (real)')

    z = np.zeros((2,2),'complex')
    for i in range(2):
        for j in range(2):
            abs_z  = np.sqrt(5 * freq * rho[i,j])
            z[i,j] = cmath.rect(abs_z ,math.radians(phi[i,j]))

    return z 


def compute_determinant_error(z_array, z_err_array, method = 'theoretical', repeats=1000):
    """
    compute the error of the determinant of z using a stochastic method
    seed random z arrays with a normal distribution around the input array

    :param z_array: z (impedance) array containing real and imaginary values
    :param z_err_array: impedance error array containing real values,
                        in MT we assume the real and imag errors are the same
    :param method: method to use, theoretical calculation or stochastic
    
    :return: error: array of real values with same shape as z_err_array 
                    representing the error in the determinant of Z
    :return: error_sqrt: array of real values with same shape as z_err_array 
                    representing the error in the (determinant of Z)**0.5
    
    """
    if method == 'stochatic':
        arraylist = []
        
        for r in range(repeats):
            errmag = np.random.normal(loc=0,scale=z_err_array,size=z_array.shape)
            arraylist = np.append(arraylist,z_array + errmag*(1. + 1j))
            
        arraylist = arraylist.reshape(repeats,z_array.shape[0],2,2)
        detlist = np.linalg.det(arraylist)
        
        error = np.std(detlist,axis=0)
    
    else:
        error = np.abs(z_err_array[:,0,0]*np.abs(z_array[:,1,1]) + z_err_array[:,1,1]*np.abs(z_array[:,0,0]) \
             - z_err_array[:,0,1]*np.abs(z_array[:,1,0]) - z_err_array[:,1,0]*np.abs(z_array[:,0,1]))
    
    return error




def propagate_error_polar2rect(r,r_error,phi, phi_error):
    """
        Find error estimations for the transformation from polar to cartesian coordinates.

        Uncertainties in polar representation define a section of an annulus. Find the 4 corners of this section and additionally the outer boundary point, which is defined by phi = phi0, rho = rho0 + sigma rho.
        The cartesian "box" defining the uncertainties in x,y is the outer bound around the annulus section, defined by the four outermost points. So check the four corners as well as the outer boundary edge of the section to find the extrema in x znd y. These give you the sigma_x/y. 

    """ 

    corners = [ ( np.real(cmath.rect(r-r_error, phi-phi_error)), np.imag(cmath.rect(r-r_error, phi-phi_error))),\
                 ( np.real(cmath.rect(r+r_error, phi-phi_error)), np.imag(cmath.rect(r+r_error, phi-phi_error))),\
                 ( np.real(cmath.rect(r+r_error, phi+phi_error)), np.imag(cmath.rect(r+r_error, phi+phi_error))),\
                 ( np.real(cmath.rect(r-r_error, phi+phi_error)), np.imag(cmath.rect(r-r_error, phi+phi_error))),\
                 ( np.real(cmath.rect(r+r_error, phi)), np.imag(cmath.rect(r+r_error, phi))) ]

    lo_x = [i[0] for i in corners]
    lo_y = [i[1] for i in corners]

    point =  (np.real(cmath.rect(r, phi)), np.imag(cmath.rect(r, phi)) )
    lo_xdiffs = [ abs(point[0] - i) for i in lo_x]
    lo_ydiffs = [ abs(point[1] - i) for i in lo_y]
    
    xerr = max(lo_xdiffs)
    yerr = max(lo_ydiffs)

    return xerr, yerr




def propagate_error_rect2polar(x,x_error,y, y_error):
    
    # x_error, y_error define a  rectangular uncertainty box  
    
    # rho error is the difference between the closest and furthest point of the box (w.r.t. the origin)
    # approximation: just take corners and midpoint of edges 
    lo_points = [ (x + x_error, y), (x - x_error, y), (x, y - y_error ), (x, y + y_error ),\
                  (x - x_error, y - y_error) ,(x + x_error, y - y_error) ,(x + x_error, y + y_error) ,(x - x_error, y + y_error) ]


    #check, if origin is within the box:
    origin_in_box = False
    if x_error >= np.abs(x) and y_error >= np.abs(y):
        origin_in_box = True

    lo_polar_points = [ cmath.polar(np.complex(*i)) for i in lo_points ]

    lo_rho = [i[0] for i in lo_polar_points ]
    lo_phi = [math.degrees(i[1])%360 for i in lo_polar_points ]

    rho_err = 0.5*(max(lo_rho) - min(lo_rho) )
    phi_err = 0.5*(max(lo_phi) - min(lo_phi))

    if (270 < max(lo_phi) < 360 ) and (0 < min(lo_phi) < 90 ):
        tmp1 = [ i for i in lo_phi if (0 < i < 90) ]
        tmp4 = [ i for i in lo_phi if (270 < i < 360) ]
        phi_err = 0.5*((max(tmp1) - min(tmp4))%360)


    if phi_err > 180:
        #print phi_err,' -> ',(-phi_err)%360
        phi_err = (-phi_err)%360

    if origin_in_box is True:
        #largest rho:
        rho_err = 2*rho_err + min(lo_rho)
        #maximum angle uncertainty:
        phi_err = 180.

    return rho_err, phi_err



def z_error2r_phi_error(z_real, z_imag, error):
    """
    Error estimation from rectangular to polar coordinates.
    
    By standard error propagation, relative error in resistivity is 
    2*relative error in z amplitude. 
    
    Uncertainty in phase (in degrees) is computed by defining a circle around 
    the z vector in the complex plane. The uncertainty is the absolute angle
    between the vector to (x,y) and the vector between the origin and the
    tangent to the circle.
    
    :returns:
        tuple containing relative error in resistivity, absolute error in phase
    
    :inputs:
        z_real = real component of z (real number or array)
        z_imag = imaginary component of z (real number or array)
        error = absolute error in z (real number or array)
    
    """
        
    z_amp = np.abs(z_real + 1j*z_imag)

    z_rel_err = error/z_amp
    
    res_rel_err = 2.*z_rel_err
    
    #if the relative error of the amplitude is >=100% that means that the relative 
    #error of the resistivity is 200% - that is then equivalent to an uncertainty 
    #in the phase angle of 90 degrees:
    if np.iterable(z_real):
        phi_err = np.degrees(np.arctan(z_rel_err))   
        phi_err[res_rel_err > 1.] = 90.
        
    else:
        if res_rel_err > 1.:
            phi_err = 90
        else:
            phi_err = np.degrees(np.arctan(z_rel_err))    
    
    
    return res_rel_err, phi_err
    
    

def old_z_error2r_phi_error(x,x_error,y, y_error):
    """
        Error estimation from rect to polar, but with small variation needed for 
        MT: the so called 'relative phase error' is NOT the relative phase error,
        but the ABSOLUTE uncertainty in the angle that corresponds to the relative
        error in the amplitude. 

        So, here we calculate the transformation from rect to polar coordinates, 
        esp. the absolute/length of the value. Then we find the uncertainty in 
        this length and calculate the relative error of this. The relative error of
        the resistivity will be double this value, because it's calculated by taking 
        the square of this length.
        
        The relative uncertainty in length defines a circle around (x,y) 
        (APPROXIMATION!). The uncertainty in phi is now the absolute of the 
        angle beween the vector to (x,y) and the origin-vector tangential to the
        circle.
        BUT....since the phase angle uncertainty is interpreted with regard to 
        the resistivity and not the Z-amplitude, we have to look at the square of
        the length, i.e. the relative error in question has to be halfed to get
        the correct relationship between resistivity and phase errors!!.

    """

    # x_error, y_error define a  rectangular uncertainty box  
    
    # rho error is the difference between the closest and furthest point of the box (w.r.t. the origin)
    # approximation: just take corners and midpoint of edges 
    lo_points = [ (x + x_error, y), (x - x_error, y), (x, y - y_error ), 
                    (x, y + y_error ), (x - x_error, y - y_error) ,
                    (x + x_error, y - y_error) ,(x + x_error, y + y_error) ,
                    (x - x_error, y + y_error) ]


    #check, if origin is within the box:
    origin_in_box = False
    if x_error >= np.abs(x) and y_error >= np.abs(y):
        origin_in_box = True

    lo_polar_points = [ cmath.polar(np.complex(*i)) for i in lo_points ]

    lo_rho = [i[0] for i in lo_polar_points ]
    lo_phi = [math.degrees(i[1])%360 for i in lo_polar_points ]

    #uncertainty in amplitude is defined by half the diameter of the box around x,y
    rho_err = 0.5*(max(lo_rho) - min(lo_rho) )

    rho = cmath.polar(np.complex(x,y))[0] 
    try:
        rel_error_rho = rho_err/rho
    except:
        rel_error_rho = 0.

    #if the relative error of the amplitude is >=100% that means that the relative 
    #error of the resistivity is 200% - that is then equivalent to an uncertainty 
    #in the phase angle of 90 degrees:
    if rel_error_rho > 1.:
        phi_err = 90
    else:
        phi_err = np.degrees(np.arcsin( rel_error_rho))


    return rho_err, phi_err




#rotation:
#1. rotation positive in clockwise direction
#2. orientation of new X-axis X' given by rotation angle
#3. express contents of Z/tipper (points P) in this new system (points P')
#4. rotation for points calculated as P' = ([cos , sin ],[-sin, cos]) * P <=> P' = R * P
#5. => B' = R * B and E' = R * E
# (Rt is the inverse rotation matrix)
#6. E = Z * B => Rt * E' = Z * Rt * B' => E' = (R*Z*Rt) * B' => Z' = (R*Z*Rt)  

#7. Bz = T * B => Bz = T * Rt * B' => T' = (T * Rt)
#(Since Tipper is a row vector)
#7.a for general column vectors v:  v' = R * v

# Rotation of the uncertainties:
# a) rotate Z into Z'
# b) use propagation of errors on Z' to obtain the rotated Z'err
# That is NOT the same as the rotated error matrix Zerr (although the result is similar)


def rotatematrix_incl_errors(inmatrix, angle, inmatrix_err = None) :
   
    if inmatrix is None :
        raise MTex.MTpyError_inputarguments('Matrix AND eror matrix must be defined')

    if (inmatrix_err is not None) and (inmatrix.shape != inmatrix_err.shape):
        raise MTex.MTpyError_inputarguments('Matrix and err-matrix shapes do not match: %s - %s'%(str(inmatrix.shape), str(inmatrix_err.shape)))


    try:
        degreeangle = angle % 360
    except:
        raise MTex.MTpyError_inputarguments('"Angle" must be a valid number (in degrees)')

    phi = math.radians(degreeangle)

    cphi = np.cos(phi)
    sphi = np.sin(phi)

    # JP: Changed the rotation matrix to be formulated to rotate
    # counter clockwise, I cannot find a good reason for this except that 
    # when you plot the strike and phase tensors the look correct with this
    # formulation.
    rotmat = np.array([[ cphi, sphi], [-sphi, cphi]])
    # rotmat = np.array([[ cphi, -sphi], [sphi, cphi]])
    rotated_matrix = np.dot(np.dot(rotmat, inmatrix), np.linalg.inv(rotmat))

    errmat  = None
    if (inmatrix_err is not None) :
        err_orig = np.real(inmatrix_err) 
        errmat = np.zeros_like(inmatrix_err)

        # standard propagation of errors:
        errmat[0,0] = np.sqrt( (cphi**2 * err_orig[0,0])**2 + \
                              (cphi * sphi * err_orig[0,1])**2 + \
                                  (cphi * sphi * err_orig[1,0])**2 + \
                                      (sphi**2 * err_orig[1,1])**2 )
        errmat[0,1] = np.sqrt( (cphi**2 * err_orig[0,1])**2 + \
                              (cphi * sphi * err_orig[1,1])**2 + \
                                  (cphi * sphi * err_orig[0,0])**2 + \
                                      (sphi**2 * err_orig[1,0])**2 )
        errmat[1,0] = np.sqrt( (cphi**2 * err_orig[1,0])**2 + \
                              (cphi * sphi * err_orig[1,1])**2 +\
                                  (cphi * sphi * err_orig[0,0])**2 + \
                                      (sphi**2 * err_orig[0,1])**2 )
        errmat[1,1] = np.sqrt( (cphi**2 * err_orig[1,1])**2 + \
                              (cphi * sphi * err_orig[0,1])**2 + \
                                  (cphi * sphi * err_orig[1,0])**2 + \
                                      (sphi**2 * err_orig[0,0])**2 )

    return rotated_matrix, errmat


def rotatevector_incl_errors(invector, angle, invector_err = None):
    #check for row or column vector 
    
    if invector is None :
        raise MTex.MTpyError_inputarguments('Vector AND error-vector must be defined')

    if (invector_err is not None) and (invector.shape != invector_err.shape):
        raise MTex.MTpyError_inputarguments('Vector and errror-vector shapes do not match: %s - %s'%(str(invector.shape), str(invector_err.shape)))

    try:
        degreeangle = angle%360
    except:
        raise MTex.MTpyError_inputarguments('"Angle" must be a valid number (in degrees)')

    phi = math.radians(degreeangle)
    
    cphi = np.cos(phi)
    sphi = np.sin(phi)

    # JP: Changed the rotation matrix to be formulated to rotate
    # counter clockwise, I cannot find a good reason for this except that 
    # when you plot the strike and phase tensors the look correct with this
    # formulation.
    rotmat = np.array([[ cphi, sphi],[-sphi, cphi]])
    # rotmat = np.array([[ cphi, -sphi],[sphi, cphi]])

    if invector.shape == (1, 2):
        rotated_vector = np.dot( invector, np.linalg.inv(rotmat) )
    else:
        rotated_vector = np.dot( rotmat, invector )
    
    
    errvec = None
    if (invector_err is not None) :   
        errvec = np.zeros_like(invector_err)

        if invector_err.shape == (1, 2):
            errvec = np.dot(invector_err, np.abs(np.linalg.inv(rotmat)))
        else:
            errvec = np.dot(np.abs(rotmat), invector_err )


    return rotated_vector, errvec



def multiplymatrices_incl_errors(inmatrix1, inmatrix2, inmatrix1_err = None,inmatrix2_err = None ):

    if inmatrix1 is None or inmatrix2 is None:
        raise MTex.MTpyError_inputarguments('ERROR - two 2x2 arrays needed as input')

    if inmatrix1.shape != inmatrix2.shape:
        raise MTex.MTpyError_inputarguments('ERROR - two 2x2 arrays with same dimensions needed as input')


    prod = np.array(np.dot( np.matrix(inmatrix1), np.matrix(inmatrix2)))

    if (inmatrix1_err is None) or ( inmatrix1_err is None ):
        return prod, None


    var = np.zeros((2,2))
    var[0,0] = (inmatrix1_err[0,0] * inmatrix2[0,0])**2 + (inmatrix1_err[0,1] * inmatrix2[1,0])**2+\
                (inmatrix2_err[0,0] * inmatrix1[0,0])**2 + (inmatrix2_err[1,0] * inmatrix1[0,1])**2
    var[0,1] = (inmatrix1_err[0,0] * inmatrix2[0,1])**2 + (inmatrix1_err[0,1] * inmatrix2[1,1])**2+\
                (inmatrix2_err[0,1] * inmatrix1[0,0])**2 + (inmatrix2_err[1,1] * inmatrix1[0,1])**2
    var[1,0] = (inmatrix1_err[1,0] * inmatrix2[0,0])**2 + (inmatrix1_err[1,1] * inmatrix2[1,0])**2+\
                (inmatrix2_err[0,0] * inmatrix1[1,0])**2 + (inmatrix2_err[1,0] * inmatrix1[1,1])**2
    var[1,1] = (inmatrix1_err[1,0] * inmatrix2[0,1])**2 + (inmatrix1_err[1,1] * inmatrix2[1,1])**2+\
                (inmatrix2_err[0,1] * inmatrix1[1,0])**2 + (inmatrix2_err[1,1] * inmatrix1[1,1])**2


    return prod, np.sqrt(var)



def reorient_data2D(x_values, y_values, x_sensor_angle = 0 , y_sensor_angle = 90):
    """
        Re-orient time series data of a sensor pair, which has not been in default (x=0, y=90) orientation.

        Input:
        - x-values - Numpy array
        - y-values - Numpy array
        Note: same length for both! - If not, the shorter length is taken 

        Optional:
        - Angle of the x-sensor - measured in degrees, clockwise from North (0) 
        - Angle of the y-sensor - measured in degrees, clockwise from North (0) 

        Output:
        - corrected x-values (North)
        - corrected y-values (East)
    """

    x_values = np.array(x_values)
    y_values = np.array(y_values)


    try:
        if x_values.dtype not in ['complex', 'float', 'int']:
            raise
        if len(x_values) != len(y_values):
            raise
    except:
        raise MTex.MTpyError_inputarguments('ERROR - both input arrays must be of same length')

    if len(x_values) != len(y_values):
        l = min(len(x_values) , len(y_values))
        x_values = x_values[:l]
        y_values = y_values[:l]

    in_array = np.zeros((len(x_values), 2), x_values.dtype)

    in_array[:,0] = x_values
    in_array[:,1] = y_values

    try:
        x_angle = math.radians(x_sensor_angle)
        y_angle = math.radians(y_sensor_angle)
    except:
        raise MTex.MTpyError_inputarguments('ERROR - both angles must be of type int or float')
       

    T = np.matrix( [[ np.real(cmath.rect(1,x_angle)), np.imag(cmath.rect(1,x_angle))],[np.real(cmath.rect(1,y_angle)), np.imag(cmath.rect(1,y_angle))]])

    try:
        new_array = np.dot(in_array, T.I)
    except:
        raise MTex.MTpyError_inputarguments('ERROR - angles must define independent axes to span 2D')

    #print new_array.shape

    return new_array[:,0], new_array[:,1]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Sep  6 19:15:17 2020

@author: Daniel03
"""

import os 
from tkinter import *
from gui_avgapp import Gui_Ediwrite

#=============================


class Press_EDIR(Frame):
    
    def __init__(self, master =None, coul='grey'):
        Frame.__init__(self)
        
        Label(self,text="Path to edifiles:").grid(row=0, column=0, sticky=W)
        self.input_path=Entry(self, width=70)
        self.input_path.grid(row=0,column=1,padx=5, pady=5)
        
        Label(self,text="Path to Coordinates File:").grid(row=0, column=2,sticky=W)
        self.input_blnpath=Entry(self,width=70)
        self.input_blnpath.grid(row=0, column=3,padx=10, pady=5)
        
        
        Label(self,text="Coordinates Filename:").grid(row=1, column=0,sticky=W)
        self.input_blnFile=Entry(self,width=70)
        self.input_blnFile.grid(row=1, column=1,padx=10, pady=5)
        
        Label(self,text="savepath:").grid(row=1, column=2,sticky=W)
        self.input_savepath=Entry(self,width=70)
        self.input_savepath.grid(row=1, column=3,padx=10, pady=5)
        
        
        can = Canvas(self, bg ="white", width =1000, height =300)
        can.grid(row =2, column =1, columnspan =3)
        
        #create Frame 
        bou1=Button(self, text="REWRITE", command=self.rewrite_edi)
        bou1.grid(row=3, column=1, columnspan=2, sticky=W, padx=10, pady=5)
        # bou2=Button(self,text="QUIT", command=self.root.quit)
        # bou2.grid(row=3 ,column=3, sticky=E, padx=10, pady=5)
        
        # self.root.mainloop()
        # self.root.destroy()
        self.master.bind("<control-Z>", self.rewrite_edi)
        self.master.title("EDIREWRITE")
        self.pack()


    def rewrite_edi(self):
        Gui_Ediwrite(edipath=self.input_path.get(),
                          bln_coordinates_filepath=self.input_blnpath.get(),
                          coordinateFile=self.input_blnFile.get(),
                          savepath=self.input_savepath.get())

if __name__=="__main__":
    fen=Tk()
    ff=Press_EDIR(master=fen)
    ff.mainloop()

    # fn=Press_STND()






    


# -*- coding: utf-8 -*-
"""
    .Console scripts. 
    
    $ python pycsamt occam2d_build_in 
    
    write  occam2d input files with edifiles.
    expected files are ('OccamDataFile.dat', 'Occam2DMesh', 'Occam2DModel', 
                        'ITERxx.iter', 'Startup')
    `xx` is= number of iteration.
    
Created on Tue Apr  6 20:53:41 2021

@author: @Daniel03
"""
from  pycsamt.modeling.occam2d import occam2d_write
from pycsamt.gui.wrap_console_scripts import wrap_cscripts as wrs

default_kwargs = {'Path to edifiles - str - ' : None , 
                  'Interpolate frequencies  - bool- '  : False, 
                  'Interpolate frequencies range (freqmin, freqmax, number of frequency) - tuple - ':(-1, 4, 17),
                  'Number of model layers*  - int - ': 31,
                  'Investigation depth (Z target) in meter** - float - ' :1100., 
                  'Z bottom in meter**  - float - ': 5000.,
                  'First layer thickness  in meter**  - float - ': 5.,
                  'Start model resistivity in log10**  - float -':1.,
                  'Number of iteration to run * - int -': 100 , 
                  'Geoelectric strike in degree E-N** - float - ': 0.,
                  'Occam2d model mode - str -' :'6',
                  'TM phase error in % - float** -': 20., 
                  'TM resistivity error in % **- float -': 10., 
                  'TE phase error in % **- float -': 20., 
                  'TE resistivity error in %** - float -': 10., 
                  'Model cell width in meter **- float - ':5, 
                  'Horizontal x pad multiplier** -float -':1.7, 
                  'Brick trigger **- float -': 1.12, 
                  
                  }
      
def main(): 
    
    wrs_obj = wrs(default_kwargs =default_kwargs)
    for kwey, kwvalues in wrs_obj.sanitize_kwargs.items():
        if kwey.find('Path to edifiles') >=0 : edipath = kwvalues
        if kwey.find('frequencies  - bool')>=0 : interpolate_frequency = kwvalues
        if kwey.find('frequencies range') >= 0 : frequency_interpolate_range = kwvalues 
        if kwey.find('of model layers') >= 0 : number_of_model_layers = kwvalues 
        if kwey.find('depth (z target)') >=0: expected_investigation_depth_or_z_target = kwvalues
        if kwey.find('Z bottom in')>=0 : exaggerate_vertical_z_bottom  = kwvalues
        if kwey.find('First layer thickness') >=0 :model_first_layer_thickness_z1 = kwvalues
        if kwey.find('Start model resistivity') >= 0 : starting_res_model_value= kwvalues 
        if kwey.find('iteration to run') >= 0 : number_of_iteration_to_run= kwvalues 
        if kwey.find('strike in degree') >= 0 : geoelectric_strike= kwvalues 
        if kwey.find('Occam2d model mode') >=0: occam_mode= kwvalues
        if kwey.find('TM phase')>=0 : TM_phase_error = kwvalues
        if kwey.find('TM resistivity') >=0: TM_res_error= kwvalues
        # if kwey.find('TE phase')>=0 : path2 = kwvalues
        # if kwey.find('TE resistivity') >=0: iter2dat_fn= kwvalues
        if kwey.find('cell width in')>=0 : model_cell_width = kwvalues
        if kwey.find('x pad multiplier')>=0 : horizontal_node_x_pad_multiplier = kwvalues
        if kwey.find('Brick trigger  ')>=0 : brick_trigger = kwvalues

  

    occam2d_write.buildingInputfiles(edi_fn =edipath,
                               geoelectric_strike= geoelectric_strike,
                               interpolate_freq= interpolate_frequency, 
                               intp_freq_logspace =frequency_interpolate_range, 
                                iteration_to_run= number_of_iteration_to_run, 
                                resistivity_start = starting_res_model_value, 
                                res_tm_err= TM_res_error, 
                                phase_tm_err= TM_phase_error,
                                occam_mode= occam_mode, 
                                n_layers = number_of_model_layers , 
                                cell_width = model_cell_width , 
                                x_pad_multiplier = horizontal_node_x_pad_multiplier, 
                                trigger =brick_trigger, 
                                z_bottom =exaggerate_vertical_z_bottom , 
                                z1_layer =model_first_layer_thickness_z1, 
                                z_target = expected_investigation_depth_or_z_target, 
                            )
if __name__ == '__main__': 
    main()


# -*- coding: utf-8 -*-
"""
Created on Tue Apr  6 12:58:45 2021
    .Console scripts. 
    write pseudo-wrapper  to interact with python shell environnement. 
    .. note:: Try to use dict_kwargs to collect input arguments values, 
            as well as defaults values .
            
@author: @Daniel03

"""

from pycsamt.utils._csamtpylog import csamtpylog 

_logger =csamtpylog.get_csamtpy_logger(__name__)


class wrap_cscripts(object): 
    """
    get input arguments from user , sanitize it and convoy towards required 
    modules 
    :param default_kwargs: dictionnary of kwargs arguments with the defaults 
        values 
    :type default_kwargs: dict 
    
    .. note:: to sanitize  input arguments , provided `*` for arguments which
        must be converte into `integer` and double `**` for those  who should be 
        convert into `float`. If conversion is not possible, will
        take the default arguments
        
    :Example :
        >>> from pycsamt.gui import wrap_console_scripts as wrs
        >>>  indico ={'Path to avg file':os.path.join(os.path.abspath('.') ,
        ...                                              'data', 'avg') 
        ...     'number of skin depth': 9,'Reference frequency':1024. }
        >>> indico = None 
        >>> cs_obj = wrs.wrap_cscripts(default_kwargs = default_kwargs, 
        ...                   input_kwargs =indico)
        >>> print(cs_obj.sanitize_kwargs)
    """
    
    def __init__(self, default_kwargs, **kws) :
        
        self._logging = csamtpylog.get_csamtpy_logger(self.__class__.__name__)
        
        self.default_kwargs =default_kwargs
  
        self.input_kwargs =kws.pop('input_kwargs', {})
        if self.input_kwargs is None : self.input_kwargs={}
        
        self.sanitize_kwargs ={}

        if self.default_kwargs is not None : 
            self._console_scripts()
        
        
    def _console_scripts(self, default_kwargs=None, **kws):
        """
        Console scripts methods used to collect input values for sanitizeed to 
        the corresponding functions before calling main function. 

        :param default_kwargs: default values acceptables for the func
        :type default_kwargs: dict
        :param input_kwargs: dictionnary of input values . If values of dict is 
            given,values should be considered as consoles input  values and will
            not request again. 
        :type input_kwargs: dict 
        
        :return: sanitized input request and values 
        :rtype: dict

        """
        self._logging.info('wraps input arguments value from user ! ')

        if default_kwargs is not None : 
            self.default_kwargs =default_kwargs 
            
        input_kwargs =kws.pop('input_kwargs', None)
        
        if input_kwargs is not None :
            if not isinstance(input_kwargs , dict):
                self.input_kwargs ={}
            else : 
                self.input_kwargs = input_kwargs 
      
        for kk, (knames, values) in enumerate(list(self.default_kwargs.items())):

            if len(list(self.input_kwargs.keys())) !=0 : 
                tem_= list(self.input_kwargs.keys())
                for jj, iname in enumerate(tem_): 
                    if knames.lower().find(iname.lower()) >=0 : #value found 
                        user_response = self.input_kwargs[iname]
                        break 
                    else : user_response = None  # value no found then ask the user 
                if user_response is None : #  request the user   
                    user_response = input('> {0} [{1}]: '.format(knames, values))
                    if user_response in ['','\n' , ' ']:
                        user_response = values
                        print('--> set to default = {}'.format(values))

            elif len(list(self.input_kwargs.keys())) ==0 :

                user_response = input('> {0} [{1}]: '.format(knames, values))
                if user_response in ['','\n' , ' ']:
                    user_response = values
                    print('--> set to default = {}'.format(values))
                
            if user_response =='None': user_response =None 
            litteral_txt = knames + ' [{0}]'.format(values)
            
            if '*' or '**' in litteral_txt : 
                litteral_text , user_response = self._sanitize_user_response(
                    litteral_text =litteral_txt, user_response=user_response,
                    default_value = values)
                
            self.sanitize_kwargs[litteral_text]= user_response 
            
    @staticmethod   
    def _sanitize_user_response(litteral_text, user_response,  default_value):
        """
        Sanitize user response with default types. In fact recheck string 
        input values from user . Then figure out  value with its 
        corresponding type using  jokers `**` and `*`. 
            * `**` means ---> convert into float values 
            * `*` means convert into interger values .
            
        """
        _logger.info(' pipeline to sanitize input argument !')
    
        dc =None 
        if  litteral_text.find('*')>=0 : dc = '*'
        elif litteral_text.find('**')>=0 : dc ='**'
        
        if dc is not None : 
            try : 
                if dc =='*' : user_response = int(user_response) 
                elif dc=='**' :user_response = float(user_response)
            
            except : 
                print(' Wrong given value ! '
                          ' should set to default = {0}'.format(default_value))
                user_response =default_value 
                
            litteral_text= litteral_text.replace(dc, '')     

        return litteral_text , user_response 



import functools
import inspect
import os
from pycsamt.utils._csamtpylog import csamtpylog


class deprecated(object):
    """
        Description:
            used to mark functions, methods and classes deprecated, and prints warning message when it called
            decorators based on https://stackoverflow.com/a/40301488

        Usage:
            todo: write usage

        Author: YingzhiGou
        Date: 20/06/2017
    """
    def __init__(self, reason):  # pragma: no cover
        if inspect.isclass(reason) or inspect.isfunction(reason):
            raise TypeError("Reason for deprecation must be supplied")
        self.reason = reason

    def __call__(self, cls_or_func):  # pragma: no cover
        if inspect.isfunction(cls_or_func):
            if hasattr(cls_or_func, 'func_code'):
                _code = cls_or_func.__code__
            else:
                _code = cls_or_func.__code__
            fmt = "Call to deprecated function or method {name} ({reason})."
            filename = _code.co_filename
            lineno = _code.co_firstlineno + 1

        elif inspect.isclass(cls_or_func):
            fmt = "Call to deprecated class {name} ({reason})."
            filename = cls_or_func.__module__
            lineno = 1

        else:
            raise TypeError(type(cls_or_func))

        msg = fmt.format(name=cls_or_func.__name__, reason=self.reason)

        @functools.wraps(cls_or_func)
        def new_func(*args, **kwargs):  # pragma: no cover
            import warnings
            warnings.simplefilter('always', DeprecationWarning)  # turn off filter
            warnings.warn_explicit(msg, category=DeprecationWarning, filename=filename, lineno=lineno)
            warnings.simplefilter('default', DeprecationWarning)  # reset filter
            return cls_or_func(*args, **kwargs)

        return new_func


class gdal_data_check(object):
    _has_checked = False
    _gdal_data_found = False
    _logger = csamtpylog.get_csamtpy_logger(__name__)

    def __init__(self, func, raise_error=False):
        """
        this decorator should only be used for the function that requres gdal and gdal-data
        to function correctly.

        the decorator will check if the GDAL_DATA is set and the path in GDAL_DATA is exist.
        If GDAL_DATA is not set, then try to use external program "gdal-config --datadir" to
        findout where the data files are installed.

        If failed to find the data file, then ImportError will be raised.

        :param func: function to be decorated
        """
        self._func = func
        if not self._has_checked:
            self._gdal_data_found = self._check_gdal_data()
            self._has_checked = True
        if not self._gdal_data_found:
            if(raise_error):
                raise ImportError("GDAL  is NOT installed correctly")
            else:
                print ("Ignore GDAL as it is not working. Will use pyproj")

    def __call__(self, *args, **kwargs):  # pragma: no cover
        return self._func(*args, **kwargs)

    def _check_gdal_data(self):
        if 'GDAL_DATA' not in os.environ:
            # gdal data not defined, try to define
            from subprocess import Popen, PIPE
            self._logger.warning("GDAL_DATA environment variable is not set  Please see https://trac.osgeo.org/gdal/wiki/FAQInstallationAndBuilding#HowtosetGDAL_DATAvariable ")
            try:
                # try to find out gdal_data path using gdal-config
                self._logger.info("Trying to find gdal-data path ...")
                process = Popen(['gdal-config', '--datadir'], stdout=PIPE)
                (output, err) = process.communicate()
                exit_code = process.wait()
                output = output.strip()
                if exit_code == 0 and os.path.exists(output):
                    os.environ['GDAL_DATA'] = output
                    self._logger.info("Found gdal-data path: {}".format(output))
                    return True
                else:
                    self._logger.error(
                        "\tCannot find gdal-data path. Please find the gdal-data path of your installation and set it to "
                        "\"GDAL_DATA\" environment variable. Please see "
                        "https://trac.osgeo.org/gdal/wiki/FAQInstallationAndBuilding#HowtosetGDAL_DATAvariable for "
                        "more information.")
                    return False
            except Exception:
                return False
        else:
            if os.path.exists(os.environ['GDAL_DATA']):
                self._logger.info("GDAL_DATA is set to: {}".format(os.environ['GDAL_DATA']))

                try:
                    from osgeo import osr
                    from osgeo.ogr import OGRERR_NONE
                except:
                    self._logger.error("Failed to load module osgeo; looks like GDAL is NOT working")
                    # print ("Failed to load module osgeo !!! ")

                    return False
                # end try

                return True
            else:
                self._logger.error("GDAL_DATA is set to: {}, but the path does not exist.".format(os.environ['GDAL_DATA']))
                return False

class redirect_cls_or_func(object) :
    """
        Description:
            used to redirected functions or classes. Deprecated functions or class 
            can call others use functions or classes.
            
        Usage:
            .. todo:: use new function or class to replace old function method or class
                with multiple parameters.

        Author: @Daniel03
        Date: 18/10/2020
    """
    
    _logger = csamtpylog.get_csamtpy_logger(__name__)
    
    def __init__(self, *args, **kwargs) :
        """
        self.new_func_or_cls is just a message of deprecating 
        warning . It could be a name of new function  to let user 
        tracking its code everytime he needs . 

        """
        
        self._reason=[func_or_reason for func_or_reason in args if type(func_or_reason)==str][0]
        if self._reason is None :
            
            raise TypeError(" Redirected reason must be supplied")
        

        self._new_func_or_cls = [func_or_reason for func_or_reason in \
                                 args if type(func_or_reason)!=str][0]

        if self._new_func_or_cls is None:
            raise Exception(" At least one argument must be a func_method_or class."
                            "\but it's %s."%type(self._new_func_or_cls))
            self._logger.warn("\t first input argument argument must be a func_method_or class."
                            "\but it's %s."%type(self._new_func_or_cls))
            

    def __call__(self, cls_or_func)  : #pragma :no cover

        if inspect.isfunction(self._new_func_or_cls) : 
            if hasattr(self._new_func_or_cls, 'func_code'):
                _code =self._new_func_or_cls.__code__
                lineno=_code.co_firstlineno+1
            else :
                # do it once the method is decorated method like staticmethods
                try:
                    _code =self._new_func_or_cls.__code__ 
                except : 
                    pass

            lineno=self._new_func_or_cls.__code__.co_firstlineno
            
            fmt="redirected decorated func/methods .<{reason}> "\
                "see line {lineno}."
            
        elif inspect.isclass(self._new_func_or_cls): 
            _code=self._new_func_or_cls.__module__
            # filename=os.path.basename(_code.co_filename)
            lineno= 1
            
            fmt="redirected decorated class :<{reason}> "\
                "see line {lineno}."
        else :
            # lineno=cls_or_func.__code__.co_firstlineno
            lineno= inspect.getframeinfo(inspect.currentframe())[1]
            fmt="redirected decorated method :<{reason}> "\
                "see line {lineno}."
        
        msg=fmt.format(reason = self._reason, lineno=lineno)
        # print(msg)
        self._logger.info(msg)
            #count variables : func.__code__.co_argscounts
            #find variables in function : func.__code__.co_varnames
        @functools.wraps(cls_or_func)
        def new_func (*args, **kwargs):
            
            return cls_or_func(*args, **kwargs)
        return self._new_func_or_cls
    
    


            

        
            
    


# -*- coding: utf-8 -*-
"""
===============================================================================
    Copyright © 2021  Kouadio K.Laurent
    
    This file is part of pyCSAMT.
    
    pyCSAMT is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    
    pyCSAMT is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public License
    along with pyCSAMT.  If not, see <https://www.gnu.org/licenses/>.
===============================================================================  

.. _module-GeoDataBase::`pycsamt.geodrill.geoDB.sql_recorder`
 
    :synopsis: Special class to manage outputs-input requests from-into 
                SQL  database .
                ...
                
.. warnings:: Editing this module presume that you are aware of what you  are doing.
              The module is a core of Geodrill subpackages. It is sensible module.
              However the the way the dataBase is arranged can be enhanced  and adapted
              for better convenient or other suitable purposes.
             
Created on Wed Oct 14 13:38:13 2020

@author:kkouaoLaurent alias @Daniel03
"""

# =============================================================================
import os
import sys
import numpy as np  
import pandas as pd 
import warnings
import datetime

from pycsamt.geodrill.geoCore.structural import Geo_formation 
import  pycsamt.utils.exceptions as CSex 
from pycsamt.geodrill.geoDB.dict_app import Glob
from pycsamt.geodrill.geoDB.interfaceDB import ManageDB
from  pycsamt.utils.decorator import redirect_cls_or_func
from pycsamt.utils._csamtpylog import csamtpylog 
#set LogInfos
try :
    _logger=csamtpylog.get_csamtpy_logger(__name__)
    # _logger.setLevel(logging.DEBUG)
    filename=os.path.basename(__file__)
    # lineno=__file__.__code__.co_firstlineno + 1
except Exception as error :
    _logger.warning("No basic configuration found. Try to set logging.basicConfig()"\
                    "file as .json file or .yml file",error )
    
    warnings.warn_explicit('you may configure logFile', category=DeprecationWarning,
                           filename=filename, lineno=1)
    pass 

# let set the systeme path find memory dataBase
 
sys.path.insert(0, os.path.abspath('.'))  
sys.path.insert(0, os.path.abspath('..')) 
sys.path.insert(0, os.path.abspath('../..'))  # for consistency 
sys.path.insert(0, os.path.abspath('pycsamt/geodrill/geoDB/sql_utils/sql_DB'))   

# =============================================================================

class GeoDataBase (object): 
    """
    Geodatabase class . Currently we do not create the specific pattern 
    for each geostructures. DataBase is built is built following the   
       codef `code`,   `label`,`__description`,`pattern`, `pat_size`,`pat_density`,
       `pat_thickness`,`RGBA`, `electrical_props`, `hatch`, `colorMPL`, `FGDC` .

    Arguments
    -----------
        **geo_structure_name** : str 
                Name of geological rocks , strata or layer.
                
    .. seealso:: FGDC-Digital cartographic Standard for Geological 
                Map Symbolisation. 
    
    """
    #  FGDC is not set yet , we use the  matplotlib pattern symbol makers 
    make_pattern_symbol =["/", "\\", "|", '-', '+', 'x', 'o', 'O', '.', 
                          '*', '\-', '\+', '\o', '\O', '\.', '\*'] 
    #use '\\' rather than '\'.
    # latter , i will be deprecated to FGDC geological map symbolisation . 
    
    codef = ['code','label','__description','pattern', 'pat_size',	
             'pat_density','pat_thickness','rgb','electrical_props', 
                 'hatch', 'colorMPL', 'FGDC' ]
    # geoDataBase=os.path.join(os.environ['pyCSAMT'],'pycsamt',
    #                'geodrill', 'geoDB','sql_utils', 'sql_DB', 'memory.sq3') 
    # locate the geodataBase
    geoDataBase = os.path.join(
        os.path.abspath('pycsamt/geodrill/geoDB/sql_utils/sql_DB'),'memory.sq3')
 
    # :memory: is faster we chose this options :geoDataBase.sq3 
    #in sql_DB contains drill holes and wells Tables 
    # will develop in the future extensions 

    def __init__(self, geo_structure_name=None) :
        self._logging =csamtpylog.get_csamtpy_logger(self.__class__.__name__)
        self.geo_structure_name = geo_structure_name
        self.dateTime= datetime.datetime.now().utcnow()   # Get the date time now  
        self.comment =None                                # initialise comment  text
        
  
        self._mplcolor =None 
        self._rgb =None 
        self._electrical_props=None 
        self._pattern=None

        try : 
        # to connect geodataBse 
            self.manage_geoDataBase =ManageDB(
                db_host=os.path.dirname(self.geoDataBase), 
                                      db_name ='memory.sq3')
        except : 
            mess =''.join(['Connection to geoDataBase failed! Sorry we can ',
                           'not give a suitable reply for your request!', 
                    'It would  be process with "geological structural class." !'])
            
            warnings.warn(mess)
            self.success= 0
 
        else : self.success = 1 

    def _avoid_injection (self): 
        """
        For secure , we do not firstly introduce directly the request. We will 
        check whether the object `request`   exists effectively  in our
        GeoDatabase . if not , request will be redirect to structural and 
        strata class issue from  module `structural`
        
        """
        # self.manage_geoDataBase.executeReq(" select __description  from AGS0")
        
        self.geo_structure_name=self.geo_structure_name.lower() # for consistency 
        manage = self.manage_geoDataBase.curs.execute(
            " select __description  from AGS0")
        __description = [ geoform[0] for geoform in  list(manage)]


        if self.geo_structure_name in __description : 
            self.geo_structure_exists =True 
            
        else :
            mess ='Structure <%s> does not exist  in'\
                ' our GeoDataBase yet.' % self.geo_structure_name
            
            self.geo_structure_exists =False
            warnings.warn(mess)
            self._logging.debug ('Could not find a {0} in our geoDataBase.'\
                                 ' It would be redirect to _strata and '
                                 '_structural classes for suitable processing.'.
                                 format(self.geo_structure_name))
            
            # self.manage_geoDataBase.closeDB() # close the database
            
        # if self.geo_structure_exists : 
        #     self._get_geo_structure()
        
    def _retreive_databasecolumns (self, columns): 
        """ Retreive data from database columns
        
        :param columns: Columns name is `str`. To retreive data of many columns 
                      please put the columns name on list.
        :returns: 
            list of data of each columns.
        
        :Exemple:
            >>> from pycsamt.geodriil.geoDB.sql_recorder import GeoDataBase 
            >>> dbObj = GeoDataBase()
            >>>  values = dbObj.._retreive_databasecolumns(
                    ['__description', 'electrical_props'])    
        
        """
        
        if isinstance(columns, str): 
            columns =[columns]
            
        new_columns = []
        for obj in columns : 
            if obj not in self.codef : 
                self._logging.debug(
                    f'Object `{obj}` not found in {self.codef}!'
                    'Please provide the right column name.')
                warnings.warn(f'Object `{obj}` not found in {self.codef}!'
                    'Please provide the right column name.')
            else:
                if obj =='name': obj = '__description'
                new_columns.append(obj)
            
        if len(new_columns) ==0 : # Object not found in database 
            self._logging.error('None object found in the database !')
            return 
        
        _l=[]  
        
        for obj in new_columns : 
            manage = self.manage_geoDataBase.curs.execute(
                " select %s  from AGS0"% obj)
            valuesdb = [ geoform[0] for geoform in  list(manage)]
            if len(new_columns)==1: 
                _l= valuesdb
            else: _l.append(valuesdb)
            
        self.manage_geoDataBase.closeDB() # close the database
        return _l
            
        
    def _get_geo_structure(self, structure_name=None):
        """
        After checking wether the name of structures exists , 
        let find the geoformation properties   from geodatabase  . 
      
        :param struture_name:   name of geological rock or layer 
        :type struture_name: str
        
        """
        if structure_name is not None :
            self.geo_structure_name = structure_name.lower() 
        
        if self.geo_structure_name is None : 
            warnings.warn('No name is inputted as geological formation. Sorry ,'
                          ' your request is aborted !')
            raise CSex.pyCSAMTError_SQL_manager(
                'No name is inputted as geological formation. Sorry ,'\
                 ' your request is aborted !')
                
        if self.geo_structure_name is not None :
            self._avoid_injection()
            
        if self.geo_structure_exists : 
            __geonames = list(self.manage_geoDataBase.curs.execute(
                "Select * from AGS0 where __description = '{0}'".\
                  format(self.geo_structure_name.lower())))[0]  
     
            # once value is get then set attribute 
            #for each column of geoDataBase
            for ii, codec in enumerate(self.codef) :
                if self.codef [10] == codec : 
                    self.colorMPL= __geonames[ii]
                    # if '#' in __geonames[ii] : # we assume that value
                    #colorMPL is hexadecimal value eg : #0000ff
                    self.__setattr__(codec, self.colorMPL) # color MPL format
                    # to  string tuple  like '(1.0, 0.5, 0.23)'
                    # else :# value in RGB color (0-1 --> 0 to 255 bits)

                elif self.codef [8]== codec : # set electrical properties
                # (min,max) line (1e-5,5.2e0 )
                    self.electrical_props = __geonames[ii] 
                    self.__setattr__(codec , self.electrical_props) # get 
                    #color matplotlib from property attributes

                else :  self.__setattr__(codec , __geonames[ii])
                
    def _reminder_geo_recorder(self, geo_structure_name ): 
        """
        To have reminder of geological formation into the geodatabase ,
        this method allow to output information 
        if the structure doesnt not exist, An error will occurs. 
        
        :param geo_structure_name: name of geological formation 
        :type geo_structure_name: str 
        
        """
        mess ='--->"{}" Querry  successfully executed ! '  
        
        if geo_structure_name is not None :
            self.geo_structure_name = geo_structure_name.lower() 
        
        if self.geoDataBase is not None : 
            self._avoid_injection() 
        if self.geo_structure_exists is True : 
                # keep only the tuple values 
            __geonames = list(self.manage_geoDataBase.curs.execute(
                "Select * from AGS0 where __description = '{0}'".\
                          format(self.geo_structure_name)))[0] 
            

            print(mess.format(self.geo_structure_name))
            print(__geonames)
            
        
        self.manage_geoDataBase.closeDB() # close the database 
        
                
    def _update_geo_structure (self, geo_formation_name =None, **kws): 
        """
        Update _indormation into geoDataBase . 
        
        Remember that the geodatabase is build following this table
        codef 'code','label','__description','pattern', 'pat_size',
        'pat_density','pat_thickness','rgb','electrical_props',  'hatch',
        'colorMPL', 'FGDC'. 
        
        :param geo_formation_name:  name of formation be sure the formation
                             already exists in the geoDataBase 
                             if not an error occurs 
        :type geo_formation_name: str
        
        - Update the electrical property of basement rocks = [1e99, 1e6 ]
        
        :Example:
            
            >>> from pycsamt.geodrill.geoDB.sql_recorder import GeoDataBase 
            >>> GeoDataBase()._update_geo_structure(
                **{'__description':'basement rocks', 
                    'electrical_props':[1e99, 1e6 ]})
        """
        fmt_mess = '---> {0} was successfully set to geoDataBase.'\
            ' Old value = {1} is updated to = {2}'
            
        # find geological rocks name if is in keywords dict
        if geo_formation_name is None : 
            if '__description' in list(kws.keys()) : 
                geo_formation_name=str(kws['__description'])
            elif 'name' in list(kws.keys()) : 
                geo_formation_name=str(kws['name'])
            else : 
                raise CSex.pyCSAMTError_SQL_update_geoinformation(
                    ' ! Unable to find a new geological structure name.')
    
        if not isinstance(geo_formation_name, str) : 
            raise CSex.pyCSAMTError_SQL_update_geoinformation(
                'Unacceptable rock/layer name ={0}.'\
                    ' Please provide a right rock/layer name.')
        geo_formation_name=str(geo_formation_name) # for consistency 
        
        if geo_formation_name is not None :
            self.geo_structure_name = geo_formation_name.lower() 
        
        # build new_dictionnary without the keyname and key value of dictionnay 
        tem_geodict ={geokey:geovalue for geokey , geovalue in kws.items() 
                      if not (geokey =='__description' or geokey =='name') 
                      }
        

        if self.geo_structure_name is not None :
            self._avoid_injection()
        if self.geo_structure_exists is False : 
            mess ="".join([
                ' Actually geological formation name =  <{0}> can '.format(
                    self.geo_structure_name),
                 'not be updated because', 
                ' it doesnt  exist in our DataBase. To set new geological ',
                'formation with their corresponding values ', 
                ' see { _add_geo_structure } method .'])
            self._logging.warn(mess)
            raise CSex.pyCSAMTError_SQL_update_geoinformation(
                'Update name= {0}  failed ! it doesnt not exist in '
                'geoDataBase'.format(self.geo_structure_name))
                    
            
        elif self.geo_structure_exists : # even the geostructure exists ,
        #let check wether the key provided 
            for geo_key in list (tem_geodict.keys()) : # is among the geocodes keys
                if geo_key not in self.codef : #if key provided not in geodatable keys 
                    
                    mess =''.join([
                        "Sorry the key = {0} is wrong! key doesnt".format(geo_key),
                        " exist in geoDataBase.Please provide a right ", 
                        " keys among = {0}". format(tuple(self.codef[2:]))])
                    self._logging.error(mess)
                    raise CSex.pyCSAMTError_SQL_manager(mess)
        
                elif geo_key in self.codef :
                    if geo_key.find('pat') >= 0 : 
                        try : # keep value to real
                            update_geo_values = float(kws[geo_key]) 
                        except : 
                            msg =''.join([
                                'update failed ! Could not convert',
                                ' value = {0} to float.'.format(kws[geo_key]), 
                                'Please try again later.'])
                            self._logging.error(msg)
                            raise CSex.pyCSAMTError_SQL_update_geoinformation(msg)
                            
                    # let get the formatage of all values  (properties values )  
                    elif geo_key.find('colorMPL') >=0 : 
                        self.colorMPL = kws[geo_key] 
                        update_geo_values = self.colorMPL
                    # let fill automatically the "rgb" and the colorMPL                
                    elif geo_key.find('rgb')>=0 :  
                        # keep the rgb value g : R125G90B29 and compute the colorMPL
                        self.rgb = kws[geo_key]
                        update_geo_values = self.rgb
                    elif geo_key .find('hatch') >=0 : 
                        self.hatch = kws[geo_key]
                        update_geo_values = self.hatch
                    elif geo_key.find('electrical_props') >=0 : 
                        self.electrical_props =kws[geo_key]
                        update_geo_values = self.electrical_props

                    else : update_geo_values =str (kws[geo_key])
                    # now must be put on the data base
                    # fill automaticall colorMPL when rgb is provided
                    if geo_key.find('rgb') >= 0 :  
                    
                        self.manage_geoDataBase.curs.execute(
                            "update AGS0 set rgb = '{0}'  where __description ='{1}'".
                              format( update_geo_values[0] , 
                                     self.geo_structure_name))
                        self.manage_geoDataBase.curs.execute(
                            "update AGS0 set colorMPL= '{0}'  where __description ='{1}'".
                              format(update_geo_values[1] , 
                                     self.geo_structure_name))
                    
                    else :
                        __oldvalues = list(
                            self.manage_geoDataBase.curs.execute(
                                "Select * from AGS0 where __description ='{0}'".
                                   format(self.geo_structure_name)))[0]
                        self.manage_geoDataBase.curs.execute(
                            "update AGS0 set {0}= '{1}'  where __description ='{2}'".
                               format(geo_key, update_geo_values ,
                                      self.geo_structure_name))
                        __newvalues = list(
                            self.manage_geoDataBase.curs.execute(
                                "Select * from AGS0 where __description ='{0}'".\
                                   format(self.geo_structure_name)))[0]
                    # inpout new info to database from cursor         
                    self.manage_geoDataBase.commit()    
                    
                    if geo_key.find('rgb') >=0 : 
                        print('---> {0} colors was successfully set to '
                              'rgb = {1} & matplotlib rgba = {2}.'.
                              format(self.geo_structure_name,
                                     update_geo_values[0],
                                     update_geo_values[1]))
                    else : 
                        fmt_mess = '---> {0} was successfully set to'\
                            ' geoDataBase.\n ** Old value = {1} \n is '\
                                '**updated to \n New value = {2}'
                        print(fmt_mess.format(
                            self.geo_structure_name,__oldvalues, __newvalues ))
                        
            self.manage_geoDataBase.closeDB() # close the database 
    @property 
    def hatch (self):
        return self._hatch 
    @hatch.setter 
    def hatch (self, mpl_hatch):
        mm=0  # counter  of hach symbol present on the value provided 
        mpl_hatch =str(mpl_hatch) # for consitency put on string value 
        # removed the tuple sign  "( and )" if provided 
        if '('  in mpl_hatch  : mpl_hatch =mpl_hatch.replace('(', '')
        if   ')' in mpl_hatch: mpl_hatch =mpl_hatch.replace(')', '')
        
        # chech whether the value provided is right 
        if mpl_hatch == 'none' : self._hatch =mpl_hatch 
        else :
            for mpstr  in mpl_hatch : 
                if mpstr in self.make_pattern_symbol  : 
                    mm +=1 
            if len(mpl_hatch) == mm :  #all value are symbols  and put the  
                self._hatch = '(' + mpl_hatch +')'
            else : self._hatch ='none' # abandon value and initialise to None 


    @staticmethod
    def _add_geo_structure( new_geological_rock_name=None , **kws) : 
        """
        Add new _geological information  into geodatabase .
        
        DataBase properties:
        --------------------
            - code
            - label
            - __description
            - pattern 
            - pat_size	
            - pat_density
            - pat_thickness
            - rgb'
            - electrical_props
            - hatch
            - colorMPL
            - FGDC 
            
        .. note:: `__description` could be replaced by `name`.
                    `code` , `label` and `FGDC` dont need to be fill. 
                    Values are rejected if given.
    
        :param new_geological_rock_name: new name of geological formation to add 
        :type new_geological_rock_name: str 
        
        :param informations: 
            dict , must be on keyward keys  when keywords keys 
                are provided , program will check whether all keys are
                effectively the right keys. if not will aborted the process.
        :type informations: dict
        
        :Example: 
            
            >>> from pycsamt.geodrill.geoDB.sql_recorder import GeoDataBase 
            >>> geodatabase_obj= GeoDataBase._add_geo_structure( **{
            ...                                     'name': 'massive sulfure', 
            ...                                     'pattern': 218., 
            ...                                     'pat_size': 250., 
            ...                                     'pat_density': 0.75, 
            ...                                     'pat_thickness': 2., 
            ...                                     'rgb': 'R128B28',
            ...                                     'hatch': '+.+.o.+', 
            ...                                     'electrical_props':[1e0 , 1e-2],
            ...                                     } )
        """
        
        def __generate_structure_code (__description , __geocodeList) : 
            """
            Will geological description will create a code and label 

            :param __description: name of geological formation 
            :type __description: str 
  
            :returns: geological formation code 
            :rtype: str 
            
            """
            def _rev_func_code (code, CODE): 
                """
                generate code and check thin new code  doesnt not exist in 
                amongs the code of geoDataBase .

                :param code:  new_generate code 
                :type code: str 
                :param CODE: codes already exists in dataBase 
                :type CODE: str 
                
                """
                # actually the lencode is > than 3 
                mm=0
                while code  in CODE :
                    if code not in CODE : 
                        break 
                    if mm > len(code): 
                        mm=0
                    code = code + code[mm]
                    mm=mm+1
            
                return code
                
            # fisrtly code is marked by three ,main letters 
            if len(__description) == 3 :code=  __description.upper() 
            elif len(__description) > 3 : code =__description[:4].upper() 
            if len(__description) < 3 :
                nadd = 0
                code =__description # loop thin you find a code =3 
                while nadd < 2 :
                    if len(code )== 3 : 
                        break 
                    code += code[nadd]
                    nadd +=1 
                code =code.upper()
            # then check whether the new code generate exist or not.
            
            for cof in __geocodeList  : 
                if code  not in __geocodeList : return code 
                if code in __geocodeList : 
                    code =_rev_func_code(code =code , CODE=__geocodeList)  

            return code             # return new code that not exist in geocodes 
            
        _logger.info (
            'Starting process  new geological information into GeoDatabase')
        
        # find geological rocks name if is in keywords dict
        if new_geological_rock_name is None : 
            if '__description' in list(kws.keys()) : 
                new_geological_rock_name=str(kws['__description'])
            elif 'name' in list(kws.keys()) : 
                new_geological_rock_name=str(kws['name'])
            else : 
                raise CSex.pyCSAMTError_SQL_update_geoinformation(
                    ' ! Unable to find a new geo_logical structure name.')
    
        if not isinstance(new_geological_rock_name, str) : 
            raise CSex.pyCSAMTError_SQL_update_geoinformation(
                'Unacceptable rocks names ={0}.'
                 ' Please provide a right rock name.')
        new_geological_rock_name=str(new_geological_rock_name) # 
        
        
        # ---------------------------call Geodatabse --------------------------
        
        geoDataBase_obj =  GeoDataBase(new_geological_rock_name)
        
        # initialise to 'none' value 
        mmgeo={geokey : 'none' for geokey in geoDataBase_obj.codef } 
        
        if geoDataBase_obj.success ==1 :  geoDataBase_obj._avoid_injection()
        elif geoDataBase_obj.success  ==0:
            mess = "Connection to SQL geoDataBase failed ! Try again later." 
            warnings.warn(mess)
            raise CSex.pyCSAMTError_SQL_geoDataBase(mess)
            
        #----------------------------------------------------------------------
       
        if geoDataBase_obj.geo_structure_exists : 
            mess ='! Name {0} already exists in our GeoDataBase. Could not add '\
                'again as new geostructure. Use "_update_geo_geostructure method"'\
                    ' to update infos if you need !'.format(
                        geoDataBase_obj.geo_structure_name)
            warnings.warn(mess)
            _logger.error(mess)
            raise CSex.pyCSAMTError_SQL_update_geoinformation(mess)
            
        if geoDataBase_obj.geo_structure_exists is False : 
  
            # make an copy of codef useful in the case where 
            # user provided "name" as key instead of "__description" 
            import copy 
            new_codef = copy.deepcopy(geoDataBase_obj.codef)
            #  set the first value of keys to fill 
            mmgeo ['__description'] = str(new_geological_rock_name) 
            if 'name' in list(kws.keys()) : 
                 new_codef[2]= 'name'
            
            # get the list of geo_code values in DataBase 
            geoDataBase_obj.manage_geoDataBase.curs.execute(
                'Select code  from AGS0 ')
            
            __geocode =[codegeo[0] for codegeo in  
                        list(geoDataBase_obj.manage_geoDataBase.curs)]
            
            
            for key in list(kws.keys()) : 

                if key not in new_codef  : 
                    raise CSex.pyCSAMTError_SQL_update_geoinformation(
                        'Process aborted ! wrong <{0}> key!'
                         ' could not add new informations. '
                         'Please check your key !'.format(key))
                #  set code and labels from geo_description name 
                elif key  in new_codef:  
                
                    try : # check if name is provided intead of  __description 
                    # name (generaly code and label are the same)
                        mmgeo['code'] = __generate_structure_code (
                            new_geological_rock_name, __geocode)
                        mmgeo['label']= __generate_structure_code (
                            new_geological_rock_name,  __geocode)
 
                    except : # user can provide name instead of __description 
                        mmgeo['code'] = __generate_structure_code (
                            new_geological_rock_name, __geocode)
                        mmgeo['label']= __generate_structure_code (
                            new_geological_rock_name, __geocode )
 
                    
                    if key.find('pat')>= 0 :
                        geoDataBase_obj.pattern = kws[key]
                
                        for kvalue  in ['pattern', 'pat_size',
                                        't_density','pat_thickness']: 
                            if key == kvalue : 
                                mmgeo[kvalue]= kws[key]
                                
                    # set RGB value and MPL colors eg : R128G128B --.(0.50, 0.5, 1.0)
                    if key =='rgb' :
                        geoDataBase_obj.rgb= kws[key]
                        # set at the same time rgb value and color MPL 
                        mmgeo['rgb']=  geoDataBase_obj.rgb[0] 
                        mmgeo['colorMPL']=  geoDataBase_obj.rgb[1]
                        
                    # set Matplotlib color whether the rgb is not provided .
                    # if provided will skip 
                    if key=='colorMPL': 
                        if mmgeo['colorMPL'] =='none' : 
                            geoDataBase_obj.colorMPL= kws[key]
                            mmgeo['colorMPL']= geoDataBase_obj.colorMPL
                    # optional keys 
                    
                    if key == 'electrical_props' : 
                        geoDataBase_obj.electrical_props= kws[key]
                        mmgeo['electrical_props']= geoDataBase_obj.electrical_props
                    if key == 'hatch': 
                        
                        mmgeo['hatch']= str(kws[key])
                    if key == 'FGDC': 
                        mmgeo['FGDC']= str(kws[key])
                        
            # print(geoDataBase_obj.success)
            # now build insert all info 
            mm_sql=[]
            for codk in new_codef:  # build info in order and input to GeodataBase
                if codk == 'name' : codk = '__description'
                mm_sql.append(mmgeo[codk])

        
            reqSQL = 'insert into AGS0 ({0}) values ({1})'.format(
                ','.join(['{0}'.format(key) for key in geoDataBase_obj.codef ]),
                ','.join(['?' for ii in range(len(geoDataBase_obj.codef))]))
 
            try : 
                # geoDataBase_obj.manage_geoDataBase.curs.execute(reqSQL , mm_sql )
                geoDataBase_obj.manage_geoDataBase.curs.execute(
                    reqSQL , mm_sql )
                
            except : 
                mess='Process to set {0} infos failed!  Try again '\
                    'later! '.format(new_geological_rock_name)
                warnings.warn (mess)
                _logger.error(mess)
                
                raise CSex.pyCSAMTError_SQL_update_geoinformation(mess)
                
            else :
   
                geoDataBase_obj.manage_geoDataBase.commit()     
                print('---> new data ={} was successfully set into'
                      ' GeoataBase ! '.format(new_geological_rock_name)) 
            
        geoDataBase_obj.manage_geoDataBase.closeDB() # close the database 
        
                  
    @property 
    def pattern (self):
        "return geopattern"
        return self._pattern 
        
    @pattern.setter 
    def pattern (self, pattern_value):
        "configure geopattern"
        try : 
             float(pattern_value)
        except : 
            mes ='Process aborted ! Could not convert'\
                f' {pattern_value} to float number.'
            self._logging.warning(mes)
            raise CSex.pyCSAMTError_SQL_update_geoinformation(mes)

        else : 
            self._pattern = float(pattern_value)
    
    @property 
    def colorMPL(self): 
        "return geocolorMPL"
        return self._mplcolor 
    
    @colorMPL.setter 
    def colorMPL (self, mpl_color): 
        """
        configure geocolorMPL
        to set matplotlib _color in rgb value 
        value is range (0 to 1) coding to 0 to 255 bits.
        """
 
        if isinstance(mpl_color, str): # get value from database 
   
            if mpl_color.find('(')>=0 and mpl_color.find(')')>=0 :
                # build the tuple of mpl colors 
                self._mplcolor = tuple([ float(ss) for ss in 
                                         mpl_color.replace(
                                             '(', '').replace(')',
                                                              '').split(',')])
            # we assume that value colorMPL is hexadecimal value eg : #0000ff
            elif '#' in mpl_color : 
                 # assume the colorMPL is in hexadecimal 
                self._mplcolor =  str(mpl_color).lower() 
            elif 'none' in mpl_color : # initilisation  value 
                self._mplcolor =  'none' # keep the value on the dataBase 
            else : 
                import matplotlib as mpl 
                try : # try to convert color to rgba
                
                    self._mplcolor = mpl.colors.to_rgb(str ( mpl_color))
                except : 
                    raise  CSex.pyCSAMTError_SQL_manager(
                        ' Unsupported {0} color!'.format(mpl_color))  
                else :  # keep only R, G, B and abandon alpha . 
                        #Matplotlib give tuple of 4 values 
                        # as (R, G, B, alpha)
                    self._mplcolor =self._mplcolor[:3] 
                    
             # set value to database way 
        elif  isinstance(mpl_color, (list, tuple, np.ndarray)): 
            if 3 <len(mpl_color) < 3 : 
                msg =''.join(['update failed ! value = {0} '.format(mpl_color),
                              'must be a tuple of 3 values= (Red, Green, Blue)',
                             'values. Please provided a right number', 
                             '  again later.'])
                
                self._logging.error(msg)
                raise CSex.pyCSAMTError_SQL_update_geoinformation(msg)
            # let check whether the value provided can be converted to float    
            if len(mpl_color)==3 :  
                try : 
                     self._mplcolor= tuple( [float(ss) for
                                             ss in list( mpl_color)])
                except : 
                     msg =''.join(['update failed ! Could not convert value ',
                                   '= {0} to float.'.format(mpl_color), 
                                   'Please try again later.'])
                     self._logging.error(msg)
                     raise CSex.pyCSAMTError_SQL_update_geoinformation(msg)
                else : 
                    # try to check if value is under 1.
                    # because clor is encoding to 1 to 255 bits 
                    for ival in  self._mplcolor: 
                        if 1 < ival <0  : 
                            if ival > 1 : fmt ='greater than 1' 
                            elif  ival <0 :
                                fmt= 'less than 0'
                            msg = ''.join([
                                'update failed ! Value provided  =',
                                ' {0} is UNacceptable value !'.format(ival),
                                f' input value is {fmt}. It must be encoding on ',
                                '1 to 255 bits as MPL colors.'])
                            raise CSex.pyCSAMTError_SQL_update_geoinformation(msg)
                            
                self._mplcolor=str( self._mplcolor) # put on str for consistency 
                
    @property 
    def rgb(self):
        "return georgb"
        return self._rgb 
    @rgb.setter 
    def rgb(self, litteral_rgb): 
        """
        configure georgb
        
        .. note:: return the rgb value and the convert rgb palette value: 
                keep the rgb value eg `R125G90B29` and compute the colorMPL
                let fill automatically the "rgb" and the colorMPL 
        """
        from pycsamt.geodrill.geoCore.structural import get_color_palette

        self._rgb=(litteral_rgb, str(
            get_color_palette(RGB_color_palette=litteral_rgb)))
        
    @property 
    def electrical_props(self): 
        "return electrical property"
        return self._electrical_props
    
    @electrical_props.setter 
    def electrical_props(self,range_of_rocks_resvalues):
        """
        configure electrical property
        
        .. note:: Electrical_property of rocks must a tuple of resisvity ,
                 max and Min bounds  eg : [2.36e-13, 2.36e-3]
        """
        # electrical props were initialse by float 0. 
        if isinstance(range_of_rocks_resvalues , str) : 
            if '(' in range_of_rocks_resvalues  : 
                self._electrical_props = tuple([ 
                    float(ss) for ss in # build the tuple of mpl colors 
                    range_of_rocks_resvalues .replace('(',
                                                '').replace(')',
                                                            '').split(',')]) 
            elif 'none' in range_of_rocks_resvalues : 
                self._electrical_props =.0
                
        elif isinstance(range_of_rocks_resvalues,(list,tuple, np.ndarray)): 
            if len(range_of_rocks_resvalues) ==2  : 
                try : 
                    self._electrical_props =[float(res) 
                                       for res in range_of_rocks_resvalues]
                except : 
                    raise CSex.pyCSAMTError_SQL_update_geoinformation(
                        ' !Could not convert input values to float.')
                else :# range the values to min to max 
                    self._electrical_props =sorted(self._electrical_props) 
                    self._electrical_props = str(tuple(self._electrical_props))
            else : 
                # force program to format value to 0. float 
                self._electrical_props = .0   
                    
        elif not isinstance(range_of_rocks_resvalues,
                            (list,tuple, np.ndarray)) or len(
                                range_of_rocks_resvalues)!=2:
            try : # 0 at initialization
                range_of_rocks_resvalues = float(range_of_rocks_resvalues)  
            except  : 
                if len(range_of_rocks_resvalues) > 1: fmt ='are'
                else :fmt ='is'
    
                mess = ''.join([
                    'Unable to set electrical property of rocks.', 
                    ' We need only minimum and maximum resistivities',
                    ' bounds. {0} {1} given'.format(
                        len(range_of_rocks_resvalues), fmt)])
                        
                self._logging.error(mess)
                warnings.warn(mess)
                raise CSex.pyCSAMTError_SQL_update_geoinformation(mess)
            else : 
                
                self._electrical_props = .0 #mean value  initialised 
                

    @property
    def _setGeoDatabase(self): 
        """
        .. note:: property of  GeoDataBase -create the GeoDataBase
                 Setting geoDataBase table
                 No Need to reset the DataBase at least you dropped the table,
                 avoid to do that if you are not sure of what you are doing. 
        """
        import  pycsamt.utils.func_utils as func 
        #set other attributes
        # create connection 
        # try : 
        #     # call manage DB  so to connect geodataBse 
        #     manage_geoDataBase =ManageDB(db_host=os.path.dirname(self.geoDataBase), 
        #                                  db_name ='memory.sq3')
        # except : pass 
        # if success is True  : 
        # createa tABLE IN THE DATABASE
        req  = ''.join(['create table AGS0', 
                      ' (', 
                      '{0} TEXT,', 
                       ' {1} TEXT,', 
                       ' {2} TEXT,', 
                       ' {3} REAL,', 
                       ' {4} REAL,', 
                       ' {5} REAL,', 
                       ' {6} REAL,', 
                       ' {7} TEXT,', 
                       ' {8} REAL,', 
                       ' {9} TEXT,', 
                       ' {10} TEXT,', 
                       ' {11} TEXT', 
                       ')'
                       ]) 
        
        # create Request to insert value into table 
        mes ='insert into AGS0 ('+ ','.join(['{}'.format(icode)
                                             for icode in self.codef])
        enter_req = mes+ ')'
        

        # create  Geoformation objets 
        geo_formation_obj =Geo_formation()      # set oBject of geostructures 
        
        new_codef ,  new_codef[-1], new_codef[0]= self.codef [:8], 'color', 'codes'
        # get attribute from geoformation and #build new geo_formations _codes 
        geo_form_codes =[ getattr(geo_formation_obj, codehead) 
                         for codehead in new_codef ]
        geo_form_codes =func.concat_array_from_list(geo_form_codes,
                                                    concat_axis=1)
        # generate other main columns of tables to fill laters 
        geo_add_form_codes = func.concat_array_from_list(
            list_of_array = [ np.zeros((len(geo_formation_obj.codes),)), 
                            np.full((len(geo_formation_obj.codes),), 'none'), 
                            np.array([str (clsp) 
                                      for clsp in geo_formation_obj.mpl_colorsp]),
                            np.full((len(geo_formation_obj.codes),),'none'),
                                                                      ],
                                                         concat_axis=1 )
        # create Table resquest 
        
        req = req.format(*self.codef)              # generate a request for table creation 
        enter_req = enter_req.format(*self.codef)  # generate interrequest 

        GDB_DATA = np.concatenate((geo_form_codes,geo_add_form_codes), axis =1)
        
        # generate values Host string so to avoid injection 
        # print(req)
        values_str = 'values (' + ','.join(['?' 
                                for itg in range( GDB_DATA.shape[1])]) +')'
        insert_request = ''.join([enter_req , values_str])

        # create Table
        
        try : 
            self.manage_geoDataBase.executeReq(query=req ) 
            
        except : 
            warnings.warn('Could not create {AGS0} Table !')
            raise CSex.pyCSAMTError_SQL_geoDataBase(
                'Table AGS0 already exists !')

        if self.manage_geoDataBase.success ==1: 
            # enter the record 
            for ii, row_geoDataBase in enumerate(GDB_DATA ): 
                row_geoDataBase =tuple(row_geoDataBase)
                self.manage_geoDataBase.executeReq(
                    query=insert_request , param =row_geoDataBase ) 
        
        self.manage_geoDataBase.commit()
        self.manage_geoDataBase.closeDB()


    
class Recorder_sql(object):
    """
    Class to record data from file or pd.core.DataFrame and to tranfer 
    into SQL database. 
    
    Arguments 
    -----------
        **database** : str , 
                name of sql database 
                
        **table** : str , 
                name of table in dict_app
    
        **Glob.dicoT** : dict,
                dicoT is from dict_app module , Global class for sql variable 
                encapsulated on particular dictionnary.
                
    =========================  ==============================================
    Methods                    Description 
    =========================  ==============================================
    transferdata_to_sqlDB      transfer Data to SQL Database 
    keepDataInfos              keep informations from datafile '*csv'
    arrangeData_for_dictapp    to arrange data, acording the dict_app 
                               arangement
    =========================  ==============================================

    .. deprecated::deprecated methods 
                - recordData --> to  keepDataInfos (staticmethod)
                - set_on_dict_app --> to tarrangeData_for_dictapp(staticmethod)
    
    :Example: 
        
        >>> from pycsamt.geodrill.geoDB.sql_recorder import Recorder_sql
        >>> realpath=os.path.dirname(os.path.realpath(__file__)) 
        >>> #where 'the file'sql_recorder is located'
        >>> print(realpath)
        >>> path_to_files =os.path.normpath(os.path.join(realpath,'sql_utils',
        ...                                                 'sql_DB'))
        >>> os.chdir(path_to_files)
        >>>  filename='nofacies_data_2.csv'
        >>> memory_DB='memory.sq3'
        >>> path_to_memory=os.path.dirname(os.path.realpath(memory_DB))
        >>> Rec=Recorder_sql(database=memory_DB,table=None)
        >>> recordList1=Recorder_sql.recordData(
            data=filename,new_tablename='TEST2',sep=',')
        >>> sdico_app=Recorder_sql.set_on_dict_app( datalist=recordList0)
        >>>  sdico_app=Recorder_sql.arrangeData_for_dictapp( 
            datalist=recordList0)
        >>> trand_sql= Rec.transferdata_to_sqlDB( 
            filename=filename,record_list= None,
        ...                 table_name='exam_zju_2',comments=None,
        ...                 visualize_table_creating_query=True,
        ...                 path_to_sqlDataBase=path_to_files,
        ...                 Drop_DB_Tables='none', fetchall=True, 
        ...                 ready_to_transfer='n',
        ...                 close_connexion =False)
    """
    
    def __init__(self, database , table=None,**kwargs):
        
        self.sqlmanager=database
        self.table=table
        self.GlobDicoT=Glob.dicoT
        
        # self.descriptif =Glob.dicoT[table]
        
        
    def transferdata_to_sqlDB(self, record_list=None,
                              filename=None, table_name=None, **kwargs):
        """
        Function to transfer Data from Dict_app to SQL DataBase. Users 
        can use this function byincluding several arguments.
        The function will build the data , arrange  it and put it in 
        the dataBase by commit the dataBase. Use only this func is benefit.
        It is better to revise arguments of that function.
        
        Parameters
        ----------
            * record_list : dict, optional
                    Dictionnay build according the dict_app model. 
                    The *default* is None.
                
            *  filename : str, optional
                    file must be on ".csv" format. The default is None.
            *  table_name : str, optional
                    Name of DataBase Table. The default is None.
            
            *  comments : str  
                    little comment to identify your database table. 
                    
            *  path_to_sqlDataBase : str  
                    path where the SQL dataBase is located . 
                
            *  visualize_table_creating_query : bool 
                   If the connexion to server is unlikable  
                   set to True to see whether query entered is right or wrong.
                   
            *  Drop_DB_Tables : str, 
                   way to drop table in SQL Database . set litteral arguments like 
                  the name of database user want to drop or 
                  [ no "*" or all to drop all tables. 
                                                             
            *  Ready_to_transfer : str 
                   process to commit Database , the curso tranfered
                   the dataBase to SQL connexion.
                   set litteral 'no' or 'yes' to do.
              
            *  close_connexion : bool , 
                  set True when transfer is done . it seems connexion.close()

        Raises
        ------
            Exception occurs when Table Name is not set on dict_app.

        .. note:: The process of organization is full request of  PostgreSQL 
        
        """
        
        comments=kwargs.pop('comments', None)
        sql_DB_loc =kwargs.pop('path_to_sqlDataBase',None)
        visualize_req=kwargs.pop('visualize_table_creating_query',False)
        drop_table=kwargs.pop('Drop_DB_Tables','none')
        fetch_data=kwargs.pop('fetchall', False)
        ready_to_transfer=kwargs.pop('Ready_to_transfer', 'n')
        close_connex=kwargs.pop('close_connexion',False)
        
        if record_list == None :
            record_list=Recorder_sql.keepDataInfos(
                data=filename,new_tablename= table_name)
            
        sdico_app=Recorder_sql.arrangeData_for_dictapp(
            datalist=record_list, comments=comments)
        
        for keys , values  in sdico_app.items():
            # or keys=[keys for keys in sdico_app]--> keys =keys[0]
            #or keys = [*sdico_app]---> keys=keys[0]
            valapp,insertvalues=values
            self.GlobDicoT.__setitem__(keys, valapp)
        # name_of_sql_DATABASE=self.sqlmanager
        manDB=ManageDB(db_name=self.sqlmanager, 
                               db_host=sql_DB_loc)
        table_req=manDB.dicT_sqlDB(dicTables=self.GlobDicoT,
                         visualize_request=visualize_req)
        
        #build request for values : 

        if self.table is None : 
            self.table =table_name
            
        if self.table not in list(self.GlobDicoT.keys()):
            raise  f'No such Table <{self.table}> '\
                'in the dataBase{self.sqlmanager} ! '\
                'You may enter the right Table Name.'
            
        self.descriptif =self.GlobDicoT[self.table]
        #"INSERT INTO oeuvres(comp,  titre,  duree, interpr) VALUES(?,?,?,?)"
        value_req=table_req.replace('CREATE TABLE', 'INSERT INTO')
        
        beacons="("         # balises= "beacons" in english
        # fieldValue=[]
        # for ii, rowline in enumerate(self.descriptif):
        #     fieldValue.append(rowline[5])   # rowline of  value in dicoT.
        
        len_values =len(insertvalues[0])
        beacons = beacons+'?,'*len_values 
        beacons=beacons [:-1] +")"   # sustract the ',' and close the parenthesis 
        value_req= value_req + ' VALUES %s' % beacons
        
        # for itemvalues in insertvalues: 
        manDB.executeReq( query =value_req, param=insertvalues)
        # manDB.executeReq( query =value_req, param=None)        
        if fetch_data is True :
            manDB.print_query()
        

        if drop_table ==False or drop_table.lower() in [
                'none',"no", 'not yet','wait','not ready']:
            pass
        elif drop_table ==True or drop_table.lower() in [
                'total', 'all', 'all tables','*']:
            
            manDB.drop_TableDB( dicTables=self.GlobDicoT,drop_all=True)
            
        elif drop_table ==True and drop_table.lower() in list[
                self.GlobDicoT.keys()]:
            
            manDB.drop_TableDB( dicTables=self.GlobDicoT,
                               drop_table_name=drop_table )
        else :
            pass 
        
        if ready_to_transfer ==True or ready_to_transfer.lower() in [
                'y','alright,''yes','ok','fine',"right"] :
            manDB.commit()
                     
        if close_connex is True :
            manDB.closeDB()
            
    @staticmethod
    def keepDataInfos(data, new_tablename=None , **kwargs ):
        """
        Function to KeepData from file infos . the function 
        is otherwritten fo RecordData. The difference between two function 
        is that function organise data from each row of columns 
        
        Parameters
        ----------
            * data : str, np.array, list, or pd.core.DataFrame object
                    Data ca, be on the format above or filename of data 
                    if the argument "data" is a filename, we must be convert 
                    on ".csv" format.
                
            * new_tablename : str, optional
                    Name of database. if name is not given , the 
                    function return only list . The default is None.
 
        Raises
        ------
            IndexError.
                 if lengh of number of columns like heads of data 
                does not match the data.shape[0], then errors will occurs.  

        Returns
        -------
            list  
                list of value in the case of no name is providen for tablename.
                else  return dict if name of datatable is providen.
            
        """
        
        nump_columns=kwargs.pop('columns_of array', None)
        separate_colist=kwargs.pop('seperate_columns_list', ',')
        seperate_csvfile=kwargs.pop('sep',',')
        
        tempkeylenvalue,temp,index_item =[],[],[]
        if nump_columns is not None :
            if type (nump_columns) ==np.ndarray :
                nump_columns=nump_columns.tolist()
            elif type (nump_columns) == str :
                nump_columns=nump_columns.split(separate_colist)
            else :
                if separate_colist not in nump_columns:
                    nump_columns=[nump_columns]  # put on a list 
    
        if os.path.isfile(data) and data.endswith('.csv'): # read csv file .
            data=pd.read_csv(data, sep=seperate_csvfile,
                             delimiter=None,header='infer',
                             index_col=False)
        
        if type (data)==pd.core.frame.DataFrame:
            columns=data.columns        # get the dataframe columns
            columns=[*columns]          # unpack pd.Index on list 
            datavalues=data.to_numpy()  # recover numpy array
            
            for ii, rowlines in enumerate( datavalues) : 
                temp.append(tuple(rowlines.tolist()))
                # if ii ==0 : 
                #     for jj, item in rowlines.tolist(): 
                #         type_of_value.append(type(item))
            type_of_value=[type(item) for  item in datavalues[0,:].tolist()] 
            index_item=[columns.index(item)+1 for item in columns]
            num_of_val=[datavalues[:,ii].shape[0]
                        for ii, item in enumerate(columns)]
            tempkeylenvalue=(columns, type_of_value,
                             index_item,num_of_val,temp)
            
            if new_tablename is not None : 
                return {new_tablename:tempkeylenvalue}

            return tempkeylenvalue
            
        elif type(data) ==np.ndarray:
            
            if nump_columns is None :
                raise Exception (
                    'The primary keys is needed ! You may provide at'
                    ' least {0} fields to fill the database '
                    'like  dictionnary values first key.'.format(
                        data.shape[1]))
            if np.dim(data)==1 :
                data=data.reshape((1,data.shape[0]))
            asize=data.shape[1]         # check the number of columns
            len_nump=len(nump_columns)
            assert asize == len_nump ,'The shape at index 1 must be the same'\
                ' like the lenght << {0}>> of data you provide'.format(
                    len(nump_columns))
            # if as_key_for_DB ==None : 
            #     as_key_for_DB=['_'.join([val[:-2]for val in nump_columns])] 
                
            if asize >1 :
                for ii, rowlines in enumerate( data) : 
                    temp.append(tuple(rowlines.tolist()))
                    
                type_of_value=[type(item) for  item in data[0,:].tolist()] 
                index_item=[nump_columns.index(item)+1 
                            for item in nump_columns]
                num_of_val=[datavalues[:,ii].shape[0] 
                            for ii, item in enumerate(nump_columns)]
                tempkeylenvalue=(nump_columns,
                                 type_of_value,index_item,num_of_val,temp)
                if new_tablename is not None : 
                    return {new_tablename:tempkeylenvalue}
    
                return tempkeylenvalue
                
            elif asize ==1 :
                # type_of_value=type(data.tolist()[0])
                type_of_value=type(data.tolist())       
                if new_tablename is not None : 
                    return {new_tablename:(nump_columns[0],
                                           type_of_value,asize, data.tolist())}
                return (nump_columns[0],type_of_value,asize,
                        data.shape[1], data.shape[0],data.tolist())
            
            
    @staticmethod    
    def arrangeData_for_dictapp( datalist, **kwargs):
        """
        Function overwritten from "set_on_dictapp func". Reorganise 
        data to dict_app model.
        
        Parameters
        ----------
            * datalist : list, dict
                list of value providen for fill the dict_app.

        Raises
        ------
             pyCSAMTError_SQL_manager
                None dataname detected

        Returns
        -------
            dict
                datalist, Data arranged according to  dict_app arrangement.
        """
        
        comments=kwargs.pop('comments', None)
        dataname=kwargs.pop('name_of_table', None )
        
        build_tup=[]
        dicosqlDB ={}
        # we can populate later 
        field_dico_inv ={int:'i',bool:'b',
                         np.ndarray:'l',list:'t',
                         float:'f',str:'t',None:'n',
                         dict:'h'} # date:'date',buffer:'by'
        try : 
            from datetime import datetime , timezone
        except ImportError:
            raise 'Importation datetime module failed ! please try again'
        
        if type (datalist) is list :
            if dataname is None :
                raise 'Please set a table name !'
            elif dataname is not None :
                datalist={dataname:datalist}

        if type(datalist) is dict :

            for key, values in datalist.items():
                main_f , type_f, posi , num_of_val, value= values
                for ii, item  in enumerate(main_f) :
                    if type_f[ii] in field_dico_inv.keys():
                        type_f[ii]='{0}'.format(field_dico_inv[type_f[ii]])
                    else :
                        type_f[ii]='{0}'.format('k')
                    posi[ii]="{0:<3}".format(posi[ii])
                    num_of_val[ii]='{0:<7}'.format(num_of_val[ii])

                    # find other param 
                    id_='{0:<10}'.format(id(main_f[ii]))
                    dtime='{0:} - {1:}'.format(datetime.now(),timezone.utc)
                    
                    idkey='{0:<}'.format(key[:-2]+main_f[ii][:-2]+id_[0]+\
                                         num_of_val[ii][0]+main_f[ii][-1])
                    if comments is None :
                        comments ='{0:}{1:<7}{2}'.format(key[:int(len(key)/2)],
                                               str(id(datalist))[-7:],main_f[ii])
                    
                    build_tup.append((item, type_f[ii], id_, posi[ii],
                                     num_of_val[ii],dtime, idkey, comments))

                # build_tup.append(value) # better to use append instead of extend 
 
                    
                dicosqlDB ={key:(build_tup,value)} 
                
        return dicosqlDB 
    
    @redirect_cls_or_func( keepDataInfos,'func"recordData" was redirected '
                          'to func "keepDataInfos" on called it.')
    @staticmethod    
    def recordData (data, new_tablename=None , **kwargs ):
        """
        Func created to record data for easily set on dictionnary for sql arrangement. 
        Indeed the func displays values along each colummns and determine the type of 
        each column. This sort of arrangement may not be use for PostgreSQL.Reason why 
        the func is deprecated . Once call , it redirected to onother function above. 
        
        Parameters
        ----------
            * data : str, np.array, list,  pd.core.DataFrame object
                    Data ca, be on the format above or filename of data 
                    if the argument "data" is a filename, we must be 
                    convert on ".csv" format.
                
            * new_tablename : str, optional
                    Name of database. if name is not given , the 
                function return only list . The default is None.

        Raises
        ------
            IndexError
                 if lengh of number of columns like heads of data 
                 does not match the data.shape[0], then errors will occurs.  

        Returns
        -------
            list , dict
                list of value in the case of no name is providen for tablename.
                else return dict if name of datatable is providen.

        :Example:
            
            >>> from pycsamt.geodrill.geoDB.sql_recorder import recordData
            >>> filename='nofacies_data.csv'
            >>> recordList0=recordData(data=filename,
                                    new_tablename='example')
        """
        nump_columns=kwargs.pop('columns_of array', None)
        separate_colist=kwargs.pop('seperate_columns_list', ',')
        seperate_csvfile=kwargs.pop('sep',',')
        # table_id =kwargs.pop('id_',None)
        # comment=''
        

        tempkeylenvalue =[]
        if nump_columns is not None :
            if type (nump_columns) ==np.ndarray :
                nump_columns=nump_columns.tolist()
            elif type (nump_columns) == str :
                nump_columns=nump_columns.split(separate_colist)
            else :
                if separate_colist not in nump_columns:
                    nump_columns=[nump_columns]  # put on a list 
    
        if os.path.isfile(data) and data.endswith('.csv'): # read csv file .
            data=pd.read_csv(data, sep=seperate_csvfile,
                             delimiter=None,header='infer',
                             index_col=False)
        
        if type (data)==pd.core.frame.DataFrame:
            columns=data.columns        # get the dataframe columns 
            datavalues=data.to_numpy()  # recover numpy array
            
            for ii, kkey in enumerate( columns) : 
                # type_of_value=type(datavalues[:,ii].tolist()[0])
                type_of_value=type(datavalues[:,ii].tolist())                
                tempkeylenvalue.append((kkey,type_of_value,ii+1, 
                                        datavalues[:,ii].shape[0],
                                        datavalues[:,ii].tolist()))
            if new_tablename is not None : 
                return {new_tablename:tempkeylenvalue}

            return tempkeylenvalue
            
        elif type(data) ==np.ndarray:
            
            if nump_columns is None :
                raise Exception (
                    'The primary keys is needed ! You may provide at least'
                    ' {0} fields to fill the database '\
                                 'like  dictionnary values first key.'.format(
                                     data.shape[1]))
            if np.dim(data)==1 :
                data=data.reshape((1,data.shape[0]))
            asize=data.shape[1]         # check the number of columns
            len_nump=len(nump_columns)
            assert asize == len_nump ,'The shape at index 1 must be the same like the '\
                'lenght << {0}>> of data you provide'.format(
                    len(nump_columns))
            # if as_key_for_DB ==None : 
            #     as_key_for_DB=['_'.join([val[:-2]for val in nump_columns])] 
                
            if asize >1 :
                for ii, kkeys in enumerate(nump_columns):
                    datatolist=datavalues[:,ii].tolist()
                    # type_of_value=type(datatolist[0])
                    type_of_value=type(datatolist)                    
                    tempkeylenvalue.append((kkeys,type_of_value,
                                            ii+1,datavalues[:,ii].shape[0],
                                            datatolist))
                
                if new_tablename is not None : 
                    return {new_tablename:tempkeylenvalue}
                
                return tempkeylenvalue
                
            elif asize ==1 :
                # type_of_value=type(data.tolist()[0])
                type_of_value=type(data.tolist())       
                if new_tablename is not None : 
                    return {new_tablename:(nump_columns[0],
                                           type_of_value,asize, data.shape[1],
                                            data.tolist())}
                return (nump_columns[0],type_of_value,asize,
                        data.shape[1], data.tolist())
    
    @redirect_cls_or_func(arrangeData_for_dictapp,
                          'set_on_dict will redirect to '
                          '"arrangeData_for_dictapp once called." ')
    @staticmethod    
    def set_on_dict_app( datalist, **kwargs):
        """
        function to arrange value according to dict_app  arangement 
        The func can call python datatype to SQLdatatype . function include 
        the datetime , and the id of each data to fullfill dictionnary.
        NB : func is redirected because it's corollary of RecordData whom it also
        redirected . if you use this function , it will redirect you to 
        "arrangeData_for_dict_app".

        Parameters
        ----------
            * datalist : list, dict
                list of value providen for fill the dict_app.

        Raises
        ------
            pyCSAMTError_SQL_manager 
                DataBase no found

        Returns
        -------
            dict
                datalist , Data arranged according to  dict_app arrangement.
        """
        
        comments=kwargs.pop('comments', None)
        dataname=kwargs.pop('name_of_table', None )

        # we can populate later 
        field_dico_inv ={int:'i',bool:'b',
                         np.ndarray:'l',list:'t',
                         float:'f',str:'t',None:'n',
                         dict:'h'} # date:'date',buffer:'by'
        try : 
            from datetime import datetime , timezone
        except ImportError:
            raise 'Importation datetime module failed ! please try again'
        
        if type (datalist) is list :
            if dataname is None :
                raise 'Please set a table name !'
            elif dataname is not None :
                datalist={dataname:datalist}

        if type(datalist) is dict :

            for key, values in datalist.items():
                for ii, rowval in enumerate(values) :
                    main_f, type_f, posi, num_of_val,value = rowval
                    main_f='{0}'.format(main_f)
                    if type_f in field_dico_inv.keys():
                        type_f='{0}'.format(field_dico_inv[type_f])
                    else :
                        type_f='{0}'.format('k')
                    posi="{0:<3}".format(posi)
                    num_of_val='{0:<7}'.format(num_of_val)
                    if type(value[0])==str :
                        value=['{0:<17}'.format(elm) for elm in value]
                    # find other param 
                    id_='{0:<10}'.format(id(main_f))
                    dtime='{0:} - {1:}'.format(datetime.now(),timezone.utc)
                    
                    idkey='{0:<}'.format(key[:-2]+main_f[:-2]+\
                                         id_[0]+num_of_val[0]+main_f[-1])
                    if comments is None :
                        comments ='{0:}{1:<7}{2}'.format(key[:int(len(key)/2)],
                                               str(id(datalist))[-7:],main_f)
                    valueT=(main_f, type_f, id_, posi,
                               num_of_val,value,dtime,idkey,comments)
                    values[ii]=valueT
                    
                datalist[key]=values
                
        return datalist     

if __name__=='__main__'   : 

    dbObj = GeoDataBase()    
    values = dbObj._retreive_column(['__description', 'electrical_props'])
    print(values)
    


            
                

                
            
            
            
